import {mapValues, mapValuesAsync, nextPowerOf2} from '../util';

import MapState from '../model/state/MapState';
import MapView from './MapView';
import Renderer from './renderer/Renderer';

function loadImage(url: string): Promise<HTMLImageElement> {
	return new Promise((resolve, reject) => {
		const img = new Image();
		img.addEventListener('load', () => { resolve(img); });
		img.addEventListener('error', () => { reject(Error('failed to load ' + url)); });
		img.src = url;
	});
}

function loadImages(urls: Map<string, string>): Promise<Map<string, HTMLImageElement>> {
	return mapValuesAsync(loadImage, urls);
}

function loadTiles(tiles: Map<string, string>): Promise<Map<string, HTMLImageElement>> {
	return loadImages(mapValues((t) => `resources/sprite/tile/${t}.png`, tiles));
}

function mkTexture(gl: WebGLRenderingContext, src: HTMLCanvasElement|HTMLImageElement): WebGLTexture {
	const texture = gl.createTexture();
	if (!texture)
		throw new Error('unable to create texture');

	gl.bindTexture(gl.TEXTURE_2D, texture);
	gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, src);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
	gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);

	return texture;
}

function mkTileTexture(gl: WebGLRenderingContext, tiles: Map<string, HTMLImageElement>):
[WebGLTexture, Map<string, number>] {
	const canvas = document.createElement('canvas');
	canvas.width = nextPowerOf2(tiles.size) * MapView.tileSize;
	canvas.height = MapView.tileSize;

	let i = 0;
	const ret: Map<string, number> = new Map();
	const ctx = canvas.getContext('2d') as CanvasRenderingContext2D;

	for (const [k, tile] of tiles) {
		ctx.drawImage(tile, i * MapView.tileSize, 0);
		ret.set(k, i++);
	}

	return [mkTexture(gl, canvas), ret];
}

export async function loadMap(r: Renderer, map: MapState): Promise<MapView> {
	const tiles = await loadTiles(map.data.tiles);
	const [tileTexture, tileMap] = mkTileTexture(r.getContext(), tiles);

	return new MapView(r, map, tileTexture, tileMap);
}
